#include "u.h"
#include "../port/lib.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"
#include "ureg.h"

#include "init.h"

enum {
	ARGSIZE = 128,	/* max space for init0 arguments */ 
};

static char *av0[2] = { "/sparc64/9ultra2", 0};

/* 
 * port: Give enough context in the ureg to produce
 * a kernel stack for a sleeping process.
 * As far as I can tell this is only used when reading from devproc.
 */
void
setkernur(Ureg *xp, Proc *p)
{
	xp->pc = p->sched.pc;
	xp->sp = p->sched.sp;
	xp->r15 = (ulong)sched;
}

/* port */
void
setregisters(Ureg *xp, char *pureg, char *uva, int n)
{
	ulong pstate;

	pstate = xp->pstate;
	memmove(pureg, uva, n);
	xp->pstate = pstate;
}

/* port: set registers for exec to return to entry */
long
execregs(ulong entry, ulong ssize, ulong nargs)
{
	Ureg *ur;
	ulong *sp;

	sp = (ulong*)(USTKTOP - ssize);
	*--sp = nargs;

	ur = (Ureg*)up->dbgreg;
	ur->usp = (ulong)sp;
	ur->pc = entry - 4;	/* syscall advances it */
	return USTKTOP-BY2WD;	/* address of user-level clock */
}

static void
_dumpstack(Ureg *ur)
{
	ulong l, sl, el;
	int i;

	print("pc=%.8lux sp=%.8lux ret=%.8lux",
		ur->pc, ur->sp, ur->r15);
	if(up) {
		sl = (ulong)up->kstack;
		el = sl + KSTACK;
	} else {
		sl = 0;
		el = 0; /*XXX*/
	}
	l = (ulong)ur->sp;
	if(l > el || l < sl) {
		print("\n  out of range %lux\n", l);
		return;
	}

	print(" (stack %lux - %lux)", sl, el);
	for(i = 0; l < el && i < (4 * 20); l += 4, i++) {
		if((i&3) == 0)
			print("\n%.8lux: ", l);
		print("%.8lux ", *(ulong*)l);
	}
	print("\ndone\n");
}

/* port */
void
dumpstack(void)
{
	callwithureg(_dumpstack);
}

/* port: Return the userland PC for fault reporting and process dumping */
ulong
dbgpc(Proc *p)
{
	Ureg *ur;

	ur = p->dbgreg;
	if(ur)
		return ur->pc;
	return 0;
}

/* port */
ulong
userpc(void)
{
	return dbgpc(up);
}

/* port */
void
procsetup(Proc *p)
{
	p->fpstate = FPinit;
}

/* port */
void
procsave(Proc *p)
{
	USED(p);
	fppause();
}

/* port: call function with a copy of registers. */
void
callwithureg(void(*fn)(Ureg*))
{
	Ureg ur;

	memset(&ur, 0, sizeof ur);
	ur.sp = (ulong)&fn - BY2WD;
	// XXX ur.r2 = setSB;
	ur.r5 = (ulong)up;
	ur.r6 = (ulong)m;
	ur.r7 = (ulong)fn;
	ur.r15 = getcallerpc(&fn);
	ur.pc = (ulong)callwithureg;
	fn(&ur);
}

static void
linkproc(void)
{
	spllo();
	up->kpfun(up->kparg);
	pexit("kproc exiting", 0);
}

/* port: set kernel process to return to func with args. */
void
kprocchild(Proc *p, void (*func)(void*), void *arg)
{
	p->sched.pc = (ulong)linkproc - 8;
	p->sched.sp = (ulong)p->kstack + KSTACK;

	p->kpfun = func;
	p->kparg = arg;
}

/* kernel portion of first process */
static void
init0(void)
{
	char buf[2*KNAMELEN];

	up->nerrlab = 0;

	spllo();

	/*
	 * These are o.k. because rootinit is null.
	 * Then early kproc's will have a root and dot.
	 */
	up->slash = namec("#/", Atodir, 0, 0);
	cnameclose(up->slash->name);
	up->slash->name = newcname("/");
	up->dot = cclone(up->slash);

	kproc("alarm", alarmkproc, 0);
	chandevinit();
	if(!waserror()){
		snprint(buf, sizeof buf, "sparc64 %s", conffile);
		ksetenv("terminal", buf, 0);
		ksetenv("cputype", "sparc64", 0);
		if(cpuserver)
			ksetenv("service", "cpu", 0);
		else
			ksetenv("service", "terminal", 0);
		poperror();
	}

	/* jump out to userland! */
	touser(USTKTOP - MAXSYSARG * BY2WD - ARGSIZE);
}

static void
bootargs(ulong base)
{
	char **pav, *p;
	ulong sp;
	int i, ac, l;

	sp = base + BY2PG - MAXSYSARG * BY2WD - ARGSIZE;

	for(l = 0, ac = 0; av0[ac]; ac++)
		l += strlen(av0[ac]) + 1;
	ac ++;
	if(l + ac * BY2WD > ARGSIZE)
		panic("bootargs");

	/* 
	 * copy the argvec and strings out to the stack
	 * while relocating the pointers.
	 */
	pav = (char**)sp;
	p = (char*)(pav + ac);
	for(i = 0; av0[i]; i++) {
		pav[i] = p - base + (USTKTOP - BY2PG);
		strcpy(p, av0[i]);
		p += strlen(p);
		*p++ = 0;
	}
	pav[i] = 0;
}

/* build the first userland process */
void
userinit(void)
{
	Proc *p;
	Segment *s;
	KMap *k;
	Page *pg;

	p = newproc();
	p->pgrp = newpgrp();
	p->egrp = smalloc(sizeof *p->egrp);
	p->egrp->ref = 1;
	p->fgrp = dupfgrp(nil);
	p->rgrp = newrgrp();
	p->procmode = 0640;

	kstrdup(&eve, "");
	kstrdup(&p->text, "*init*");
	kstrdup(&p->user, eve);

	procsetup(p);

	/* kernel stack */
	p->sched.pc = (ulong)init0 - 8;	/* we jump to pc+8 */
	p->sched.sp = (ulong)p->kstack + KSTACK - MAXSYSARG*BY2WD;
	p->sched.sp &= ~7;

	/* user stack with boot arguments */
	s = newseg(SG_STACK, USTKTOP - USTKSIZE, USTKSIZE / BY2PG);
	p->seg[SSEG] = s;
	pg = newpage(1, 0, USTKTOP - BY2PG);
	segpage(s, pg);
	k = kmap(pg);
	bootargs(VA(k));
	kunmap(k);

	/* Text from initcode */
	s = newseg(SG_TEXT, UTZERO, 1);
	s->flushme ++;
	p->seg[TSEG] = s;
	pg = newpage(1, 0, UTZERO);
	memset(pg->cachectl, PG_TXTFLUSH, sizeof pg->cachectl);
	segpage(s, pg);
	k = kmap(s->map[0]->pages[0]);
	memmove((uchar*)VA(k), initcode, sizeof initcode);
	kunmap(k);

	ready(p);
}
